#! /usr/bin/python

## uttRWSegOnly.py
## Copyright (C) 2007 Ivan Uemlianin
## Contact: ivan@llaisdy.com
## 
## This file is part of Trefnydd.
##
## Trefnydd is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## Trefnydd is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Trefnydd (i.e., doc/gpl.txt); if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
## or see http://www.gnu.org/licenses/gpl.txt

## Parts of this module are based on SpeechCluster, developed at
## Canolfan Bedwyr and to which the following coyright notice applies:
##
## # Copyright (c) 2005, University of Wales, Bangor 
## # All rights reserved.
## # 
## # Redistribution and use in source and binary forms, with or without
## # modification, are permitted provided that the following conditions
## # are met:
## # 
## #   * Redistributions of source code must retain the above copyright
## # notice, this list of conditions and the following disclaimer.
## #   * Redistributions in binary form must reproduce the above copyright
## # notice, this list of conditions and the following disclaimer in the
## # documentation and/or other materials provided with the distribution.
## #   * Neither the name of the University of Wales, Bangor nor the names
## # of its contributors may be used to endorse or promote products
## # derived from this software without specific prior written permission.
## # 
## # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## # "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## # LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## # A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## # OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## # SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## # LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## # DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## # THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
## For SpeechCluster see:
## http://www.bangor.ac.uk/~cbs007/speechCluster/README.html
## http://www.bangor.ac.uk/ar/cb/wispr.php
## http://www.e-gymraeg.org/wispr/


"""
========
trefnydd
========

Package:   trefnydd
Module:    uttRWOnly
Copyright: 2007 Ivan A. Uemlianin
Contact:   ivan@llaisdy.com
Licence:   GNU GPL (see doc/gpl.txt)

uttRWOnly
=========

A version (sub/super-class?) of utterance that does only read/write operations (ie, no manipulation)

"""

# 070104: TODO: tier.name -> tier.type!


import os

from trefnydd.utils import location
from trefnydd.lib.utterance import SegmentationTier, Segment


class UttRWSegOnly(location.Located):
    sepDict = {'esps': ' ',
               'htk-lab': ' ',
               'TextGrid': '='}
    endHeadDict = {'esps': '#',
                   'htk-lab': '\n',
                   'TextGrid': '\n'}

    def __init__(self, segFn='', debug=False):
        location.Located.__init__(self, debug)

        self.readFormatDict = {
        '.esps': self.read_ESPS,
        '.lab': self.read_ESPS,
        '.seg': self.read_ESPS,
        '.htk-lab': self.read_HTKLab,
        '.htk-grm': self.read_HTKGrm,
        '.textgrid': self.read_TextGrid,
        '.pm': self.read_pm,
        '.txt': self.read_orth
        }

        self.segFn = segFn
        self.fstem = ''

        self.dataMin = 0
        self.dataMax = 0

        self.head = {} # misc metadata read from the file header
        self.tiers = []
        self.mergeDict = {}

        if segFn: self.read_format(segFn)

    def read_format(self, fn):
        fstem, fext = os.path.splitext((os.path.basename(fn)))
        if fext:
            self.fstem = fstem
            self.readFormatDict.get(fext.lower(), self.unknownFormat)(fn)
        else: # fn is a stem; load all files with same stem (inc audio)
            path, fn = os.path.split(fn)
            self.fstem = fn
            if not path: path = os.getcwd()
            for fn in [os.path.join(path, fn)
                       for fn in os.listdir(path)
                       if os.path.splitext(fn)[0] == fstem]:
                seg = UttRWSegOnly(fn)
                self.merge(seg)

    def merge(self, other):
        self.mergeDict[other.segFn] = other.head
        for tier in other.tiers:
            self.updateTiers(tier)
        if other.audioData:
            self.setAudioFn(other.getAudioFn())
            self.audioData = other.audioData
            self.audioFormat = other.audioFormat
            self.audioSampleRate = other.audioSampleRate
            self.audioSampleSize = other.audioSampleSize

    def write_format(self, format):
        if format == 'esps' or format == 'lab':
            out = self.write_ESPS(format)
        elif format == 'TextGrid':
            out = self.write_TextGrid()
        elif format in ['htk', 'htk-lab']:
            out = self.write_HTKLab()
        elif format == 'htk-grm':
            out = self.write_HTKGrm()
        return out

    def unknownFormat(self, fn):
        raise StandardError, 'Unknown File Format.'

    def save(self):
        if self.debug:
            print 'UttRWSegOnly.save(): homeDir = \n\t%s' % self.homeDir
        if self.changed:
            for tier in self.tiers:
                tier.save(self.homeDir)
            self.changed = False

    def load(self, dirn=''):
        if not dirn:
            dirn = self.homeDir
        if self.debug is True:
            print 'UttRWSegOnly.load()\n\tdirn: %s' % dirn
        for fn in os.listdir(dirn):
            if self.debug is True:
                print '\tfn: %s' % (fn)
            fullfn = os.path.join(dirn, fn)
            if fn.startswith(self.name):
                self.read_format(fullfn)
            elif os.path.splitext(fn)[1] == '.tier':
                self.updateTiers(SegmentationTier(fullfn))
    
    def getName(self):
        return self.fstem

    def setName(self, name):
        self.fstem = name
        self.changed = True

    name = property(getName, setName)

    def getHeadBody(self, fn, sep):
        data = list(open(fn))
        # lines with misc whitespace --> '\n'
        # TODO: or with misc whitespace at end of line
        for i in range(len(data)):
            if data[i].strip == '':
                data[i] = '\n'
        hidx = data.index(sep) + 1
        head, body = data[0:hidx], data[hidx:]
        return head, body

    def updateTiers(self, tier):
        tier.dataMax = max(self.dataMax, tier[-1].max)
        self.tiers.append(tier)
        maxes = [t.dataMax for t in self.tiers]
        self.dataMax = max(maxes)
        for t in self.tiers:
            t.dataMax = self.dataMax

    def getTierByName(self, tierName):
        for t in self.tiers:
            if t.name == tierName:
                return t

    def getTierByType(self, tierType):
        for t in self.tiers:
            if t.name == tierType:
                return t

##     ## idea for future format plugins
##     def ESPS_format(self, fn, action='read'):
##         if action == 'read': self.read_ESPS(fn)
##         elif action == 'write':self.write_ESPS(fn)

    def read_ESPS(self, fn):
        """
        Returns ["head", [time, something_else, label]]
        """
        try:
            head, body = self.getHeadBody(fn, '#\n')
        except: # probably htk-lab
            self.read_HTKLab(fn)
        else:
            self.parseHead(head, 'esps')
            tier = SegmentationTier()
            tier.setName('Phone') # TODO: magic text!
            if body:
                for line in body:
                    fields = line.split()
                    # fields[0] = time
                    # fields[1] = something_else  - TODO: Q: what is it?
                    # fields[2] = label
                    if len(fields) != 3:
                        if self.debug is True:
                            print 'DEBUG: N FIELDS: %s line: %s' % (fn, line)
                            print tier
                        #else:
                        pass # what then!?
                    else:
                        seg = Segment()
                        if tier:
                            seg.min = tier[-1].max
                        else: seg.min = 0
                        seg.max = eval(fields[0])
                        seg.label = fields[-1]
                        tier.append(seg)
                self.updateTiers(tier)

    def write_ESPS(self, format, tier='Phone'):
        ### TODO: magic text! only one tier!
        out = '%s\n' % (self.writeHead(format))
        # what about writing multiple tiers?
        for segment in self.getTierByName(tier):
            out = '%s\t%s\t125 %s\n' \
                  % (out, segment.max, segment.label)
        # todo: tier separator in esps format!?
        return out

    def read_HTKLab(self, fn):
        """
        Returns ["head", [start time, end time, label]]
        """
        body = list(open(fn))
        head = body.pop(0)
        
        tier = SegmentationTier()
        tier.setName('Phone') # TODO: magic text!
        if body:
            for line in body:
                fields = line.split()
                # fields[0] = start time
                # fields[1] = end time
                # fields[2] = label
                if len(fields) != 3:
                    if self.debug is True:
                        print 'DEBUG: N FIELDS: %s line: %s' % (fn, line)
                        print tier
                    #else:
                    pass # what then!?
                else:
                    seg = Segment()
                    seg.min = eval(fields[0])/10000000.0
                    seg.max = eval(fields[1])/10000000.0
                    seg.label = fields[-1]
                    tier.append(seg)
            self.updateTiers(tier)

    def write_HTKLab(self, tier='Phone'):
        ### TODO: magic text! only one tier!
        out = '%s.lab\n' % self.fstem
        # for tier in tiers:
        for segment in self.getTierByName(tier):
            out = '%s%s %s %s\n' \
                  % (out,
                     int(segment.min*10000000),
                     int(segment.max*10000000), segment.label)
        # todo: tier separator in esps format!?
        out = '%s.\n' % out
        return out

    def read_orth(self, fn):
        """
        """
        tier = SegmentationTier()
        tier.setName('Word') # TODO: magic text!
        # TODO:
        # 1: 'Word' is type not name
        # 2: Word, Phone, etc - magic text!?
        tier.parseOrthography(open(fn).read())
        self.updateTiers(tier)

    def write_orth(self, tier='Phone'):
        pass

    def read_TextGrid(self, fn):
        import re
        # TODO: doesn't read in pitch tier!
        head, body = self.getHeadBody(fn, '\n')
        self.parseHead(head, 'TextGrid')
        # the following non-indented lines can be skipped
        while body[0][0] not in [' ', '\t']: body.pop(0)
        # there follows a list of 'item [n]:'
        itemList = ''.join(body).split('item')[1:]
        for item in itemList:
            il = item.split('\n')
            #if il[1].strip() == 'class = "TextTier"':
            #    # TODO: skips pitch tiers!
            #    continue
            # if sc works withthis commented out,
            # delete it
            content = [l.strip() for l in il[1:]]
            tier = SegmentationTier()
            # first 2 lines should be class then name
            tier.head['class'] = re.search('class = "(.*?)"',
                                           content[0]).group(1)
            tier.head['name'] = re.search('name = "(.*?)"',
                                           content[1]).group(1)
            #skip till we get to the intervals
            while content and ( len(content[0]) < 12 or \
                                content[0][:10] != 'intervals ' ):
                content.pop(0)
            # each interval is 4 lines: head, xmin, xmax, text
            while content:
                interval = content[0:4]
                if len(interval) == 4:
                    seg = Segment()
                    seg.label = re.search('text = "(.*?)"',
                                          interval[3]).group(1)
                    seg.min = eval(interval[1].split('=')[1].strip())   #xmin
                    seg.max = eval(interval[2].split('=')[1].strip())   #xmax
                    tier.append(seg)
                    content = content[4:]
                else: content = []
            if len(tier):
                self.updateTiers(tier)

    def write_TextGrid(self):
        out = '%s\n' % (self.writeHead('tg'))
        out = '%sxmin = %s\n' % (out, self.dataMin)
        out = '%sxmax = %s\n' % (out, self.dataMax)
        out = '%stiers? <exists>\n' % out # assume tiers exist
        out = '%ssize = %d\n' % (out, len(self.tiers))
        out = '%sitem []: \n' % out
        for i in range(len(self.tiers)):
            tier = self.tiers[i]
            out = '%s\titem [%d]:\n' % (out, i+1)
            tierClass = tier.head.get('class', 'IntervalTier')
            out = '%s\t\tclass = "%s"\n' \
                  % (out, tierClass)
            out = '%s\t\tname = "%s"\n' \
                  % (out, tier.head.get('name', 'Labels'))
            out = '%s\t\txmin = %s\n' % (out, tier.dataMin)
            out = '%s\t\txmax = %s\n' % (out, tier.dataMax)
            if tierClass == 'IntervalTier':
                out = '%s\t\tintervals: size = %d\n' % (out, len(tier))
                for i in range(len(tier)):
                    seg = tier[i]
                    out = '%s\t\tintervals [%d]:\n\t\t\txmin = %s\n\t\t\txmax = %s\n\t\t\ttext = "%s"\n' \
                          % (out, i+1, seg.min, seg.max, seg.label)

            elif tierClass == 'TextTier':
                out = '%s\t\tpoints: size = %d\n' % (out, len(tier))
                for i in range(len(tier)):
                    seg = tier[i]
                    out = '%s\t\tpoints [%d]:\n\t\t\ttime = %s\n\t\t\tmark = "%s"\n' \
                          % (out, i+1, seg.max, seg.label)
                
        return out

    def read_pm(self, fn): #### TODO
        #Hacky version quickly
        data = list(open(fn))[8:] # discard header
        tier = SegmentationTier()
        tier.setName('Pitch')
        tier.head['class'] = 'TextTier'
        for t in [line.split()[0] for line in data]:
            seg = Segment()
            seg.min = seg.max = eval(t)
            tier.append(seg)
        self.updateTiers(tier)

    def write_pm(self):
        pass

    def read_HTKGrm(self, fn):
        pass
    
    def write_HTKGrm(self, tier='Word'):
        """HTK gram*.txt format
        $sentence = lab1 lab2 lab3 ... labn;
        ($sentence)
        """
        if len(self.tiers) > 1:
            tier = self.getTierByName(tier)
        else:
            tier = self.tiers[0]
        labList = [seg.label for seg in tier]
        labels = ' '.join(labList)
        out = '$utterance = %s;\n($utterance)\n' % (labels)
        return out


    def parseHead(self, head, format='esps'):
        """
        if header is name = value format: make dict
        else: dict keys are line nos
        """
        sep = self.sepDict[format]
        endh = self.endHeadDict[format]
        for i in range(len(head)):
            line = head[i]
            if sep in line:
                name = line.split(sep)[0]
                value = ' '.join(line.split(sep)[1:])
                self.head.setdefault(name.strip(), []).append(value.strip())
            elif line[:-1] == endh:
                return
            else: pass # what else could happen?

    def writeHead(self, format='esps'):
        out = ''
        if format == 'esps':
            order = ['signal', 'type', 'comment',
                     'separator', 'nfields']
            otherkeys = [i for i in self.head.keys() if i not in order]
            for key in order:
                for i in self.head.get(key, ['UNDEF']):
                    out = '%s%s %s\n' % (out, key, i)
            for key in otherkeys:
                for i in self.head[key]:
                    out = '%s%s %s\n' % (out, key, i)
            out = '%s#\n' % out
        elif format == 'lab':
            # hack for festival format lab files
            out = 'separator ;\nnfields 1\n#'
        elif format == 'tg':
            out = 'File type = "ooTextFile"\nObject class = "TextGrid"\n'
            # TODO: add other (nonTextGrid) header items as comments
            # TODO: Q: What is TextGrid comment marker?
        return out


def main():
    pass

if __name__ == '__main__':
    main()

#
        
